// Class MazeGUI
//
// Author: Alyce Brady
//
// License Information:
//   This class is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation.
//
//   This class is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.KeyStroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 *  Mouse in a Maze Program:<br>
 *
 *    The <code>MazeGUI</code> class provides a window in which
 *    to display a mouse in a maze program.
 *
 *  @author Alyce Brady
 *  @version 5 March 2003
 **/
public class MazeGUI extends JFrame
{
    // constants used to initialize slider
    public static final int MIN_DELAY_MSECS = 10;
    public static final int MAX_DELAY_MSECS = 1000;
    public static final int INITIAL_DELAY = MIN_DELAY_MSECS +
                                    (MAX_DELAY_MSECS-MIN_DELAY_MSECS)/2;

    // encapsulated MazeGUI instance variables
    private EnvFileChooser fileChooser = null;
    private Maze maze = null;
    private ScrollableEnvDisplay display = null;
    private int delay = INITIAL_DELAY;    // time to view the display
    private JComboBox strategyChooser = null;
    private JButton startButton = null;


  // constructor

    /** Creates a new MazeGUI with a menu bar.  
     *    @param title frame title
     *    @param backgroundColor color to paint background of environment
     *    @param viewingWidth  the width of the viewing area
     *    @param viewingHeight the height of the viewing area
     **/
    public MazeGUI(String title, Color bgColor,
                   int viewingWidth, int viewingHeight)
    {
        // Reset exception handler to notify user of exception and
        // generate trace information.
        System.setProperty("sun.awt.exception.handler",
                           GUIExceptionHandler.class.getName());
 
        // Set characteristics for the main frame.
        setTitle(title);
        setLocation(25, 15);
        setDefaultCloseOperation(EXIT_ON_CLOSE);

        // Make any appropriate menus and add to the main frame.
        setJMenuBar(makeMenus());

        // Define the contents of the main frame.
        setContentPane(makeContentPane(bgColor, viewingWidth, viewingHeight));

        pack();
        setVisible(true);
    }


  // methods that deal with constructing menus

    /** Creates the menu bar with drop-down menus on the frame.
     **/
    protected JMenuBar makeMenus()
    {
        // Make the menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Make the menu(s) to put in the menu bar.
        menuBar.add(makeFileMenu());

        return menuBar;
    }

    /** Creates a File menu.
     **/
    protected JMenu makeFileMenu()
    {
        // Make the File menu and its menu items.
        JMenu fileMenu = new JMenu("File");
        int menuMask = getToolkit().getMenuShortcutKeyMask();
        JMenuItem mItem;

        // Open menu item?

        mItem = new JMenuItem("Quit");
        mItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, menuMask));
        mItem.addActionListener( new ActionListener() { 
            public void actionPerformed(ActionEvent e) { System.exit(0); }});
        fileMenu.add(mItem);

        return fileMenu;
    }

    /** Reads a new environment from a saved data file. Gets file via
     *  file chooser dialog, reads contents using the specified file handler.
     **/
    public void openFile()
    {
        if (fileChooser == null) fileChooser = new EnvFileChooser();
        if (fileChooser.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
            return;
            
        try 
        {
            // Construct the object that knows how to read in maze information
            // from a file, and ask it to read the maze.
            MazeDataFileHandler fileHandler = new MazeDataFileHandler();
            Maze newMaze = fileHandler.readMaze(fileChooser.getSelectedFile());
            setMaze(newMaze); // this happens only if an exception was NOT thrown!
            display.showEnv();
        }
        catch (Exception ex)
        {
            JOptionPane.showMessageDialog(this,
                        "Unable to read environment from file.\n"
                        + "Reason: " + ex, "Error reading file", 
                        JOptionPane.ERROR_MESSAGE);
        }
    }


  // methods that deal with defining the main panel contents

    /** Creates the main content panel.
     *    @param backgroundColor color to paint background of environment
     *    @param viewingWidth  the width of the viewing area
     *    @param viewingHeight the height of the viewing area
     **/
    protected JPanel makeContentPane(Color bgColor,
                                    int viewingWidth, int viewingHeight)
    {
        JPanel content = new JPanel();
        content.setBorder(BorderFactory.createEmptyBorder(15, 15, 15, 15));
        content.setLayout(new BorderLayout());

        // Create a panel to display the maze and speed control slider.
        content.add(makeDisplayPanel(bgColor, viewingWidth, viewingHeight), 
                    BorderLayout.CENTER);

        // Create a panel containing control buttons.
//        content.add(makeControlPanel(), BorderLayout.WEST);

        // Enable/disable items that depend on existence of environment.
        enableItemsThatNeedEnvironment();
        return content;
    }

    /** Creates the panel for displaying an environment and a speed slider.
     *    @param backgroundColor color to paint background of environment
     *    @param viewingWidth  the width of the viewing area
     *    @param viewingHeight the height of the viewing area
     *    @return a panel with an environment display and speed slider
     **/
    protected Component makeDisplayPanel(Color bgColor,
                                         int viewingWidth, int viewingHeight)
    {
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());

        // Add the environment display and the slider to the panel.
        panel.add(makeEnvDisplay(bgColor, viewingWidth, viewingHeight),
                  BorderLayout.CENTER);

        return panel;
    }

    /** Creates a scrollable display for displaying an environment
     *  application.
     *    @param backgroundColor color to paint background of environment
     *    @param viewingWidth  the width of the viewing area
     *    @param viewingHeight the height of the viewing area
     *    @return a scrollable pane with an environment display
     **/
    protected Component makeEnvDisplay(Color bgColor,
                                       int viewingWidth, int viewingHeight)
    {
        // Create a scrollable pane with minimum cell size of 10.  There are
        // no actions defined for the graphics scrollpane, so it is always
        // "disabled."
        display = new ScrollableEnvDisplay(viewingWidth, viewingHeight,
                                           10, bgColor);
        display.setEnabled(false);

        return display;
    }

    /** Creates a speed slider for controling the speed of the animation.
     *    @return a panel containing the speed slider
     **/
    protected JPanel makeSliderPanel()
    {
        // Create a speed slider for controling the speed of the application.
        // Package the slider together with a label, and put them in a panel.
        JPanel sliderPanel = new JPanel();
        sliderPanel.setBorder(BorderFactory.createTitledBorder("Adjust Speed"));

        // Create Slow label - slider - Fast label.
        sliderPanel.add(new JLabel("Slow"));
        JSlider speedSlider = new JSlider(MIN_DELAY_MSECS, MAX_DELAY_MSECS,
                                          INITIAL_DELAY);
        speedSlider.setInverted(true);
        speedSlider.setPreferredSize(new Dimension(100, 
                                    speedSlider.getPreferredSize().height));
        speedSlider.setMaximumSize(speedSlider.getPreferredSize());
        speedSlider.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                delay = ((JSlider)evt.getSource()).getValue();
            }});
        sliderPanel.add(speedSlider);
        sliderPanel.add(new JLabel("Fast"));

        // The speed slider is always enabled.
        sliderPanel.setEnabled(true);

        return sliderPanel;
    }
    
    /** Sets the enabled status of those GUI items that need an
     *  environment to be valid. 
     **/
    public void enableItemsThatNeedEnvironment()
    {
        // The only items in this GUI that should be enabled only
        // when there is a defined maze are the Start button and
        // the tool tips for the display.
//        startButton.setEnabled(maze != null);
        if ( display != null )
            display.setToolTipsEnabled(maze != null);
    }


  // methods for building the control panel

    /** Creates the control panel.
     *    @param label description for this set of control components
     *    @return a panel containing the control components
     **/
    protected JPanel makeControlPanel()
    {
        // Create a panel for the control components (strategy chooser
        // and Start button).  Put the actual components in a smaller
        // sub-panel so that they stay at the top of the panel instead
        // of growing to fill the entire control panel, which might be
        // much longer than necessary.
        JPanel controlPanel = new JPanel();
        controlPanel.setLayout(new BorderLayout());
        String title = "Control Panel";
        controlPanel.setBorder(BorderFactory.createTitledBorder(title));

        JPanel subPanel = new JPanel();
        subPanel.setLayout(new GridLayout(2, 0));

        // Make a strategy chooser and a Start button, and add them to
        // the control sub-panel.


        controlPanel.add(subPanel, BorderLayout.NORTH);
        return controlPanel;
    }

    /** Creates the drop-down menu for choosing a movement strategy for a mouse.
     *    @param label description for this set of control components
     *    @return a panel containing the control components
     **/
    protected JComboBox makeStrategyChooser()
    {
        strategyChooser = new JComboBox();
        // how do you add items, such as a new RandomMovementStrategy,
        // to the combo box?

        return strategyChooser;
    }

    /** Gets the currently selected movement strategy for a mouse.
     *    @return a movement strategy object
     **/
    protected MovingStrategy getStrategy()
    {
        return null;
    }


  // methods that control the mouse-in-a-maze experiment

    /** Sets the Maze.
     *  (Precondition: maze is not null.)
     *    @param newMaze the maze in which to run the mouse-in-a-maze
     *                   experiment.
     **/
    public void setMaze(Maze newMaze)
    {
        maze = newMaze;
        display.setEnv(maze);
        enableItemsThatNeedEnvironment();
    }

    /** Starts the mouse-in-a-maze experiment in a parallel thread.
     **/
    public void start()
    {
        Thread myThread = new Thread()
        {
            public void run ()
            {
                // Construct a mouse with the appropriate strategy
                // at the starting location and the cheese at the
                // finishing location.  Display the initial configuration.

                // Run the experiment.

                // When done, remove the mouse.
            }
        };

        myThread.start();
    }

    /** Shows the environment.
     *  (Precondition: must have called <code>setMaze</code>.)
     **/
    public void showEnv()
    {
        display.showEnv();
        try 
        {
          Thread.sleep(delay);
        }
        catch (InterruptedException e) {}
    }

  // Nested class for handling exceptions
    
    /** Nested class that is registered as the handler for exceptions
     *  on the Swing event thread. The handler will put up an alert panel,
     *  dump the stack trace to the console, and then exit entire program 
     *  rather than persist in an inconsistent state, which would be
     *  the default behavior.
     */
    public static class GUIExceptionHandler
    {
        public void handle(Throwable e)
        {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, 
                "An error occurred. The simulation must exit." + "\nReason: " + e, 
                "Error", JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }
    }
}


